! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: array_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to process arrays in various ways.
! 
! Comments: TODO:
! 
! $Id: array_utils.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module array_utils

  use linear	! for solveLinearTriDiag()

  implicit none 

  private 

  public rebinImage
  public rebin
  public getBinNum
  public smallestIndex
  public largestIndex
  public trapezoidSum
  public vectorCross
  public maskArrayLimits
  public calcCubicSplines
  public cubicSplineInterpolate

  interface rebinImage
    module procedure rebinImageBasicReal, rebinImageWithEdgesReal
    module procedure rebinImageBasicDble, rebinImageWithEdgesDble
  end interface

  interface rebin
    module procedure rebinBasicReal, rebinWithWidthsReal, rebinWithEdgesReal
    module procedure rebinBasicDble, rebinWithWidthsDble, rebinWithEdgesDble
  end interface

  interface vectorCross
    module procedure vectorCrossReal
    module procedure vectorCrossDble
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine calcCubicSplines(x, y, loBoundType, loBoundValue, hiBoundType&
  , hiBoundValue, y2, status)
  !---------------------------------------------------------------------
  ! This is designed to work in conjunction with subroutine
  ! cubicSplineInterpolate() to provide a cubic-spline-type interpolation.
  ! The present subroutine, given an ordered series of unique x values
  ! plus corresponding y values, calculates the values at all x of the
  ! second derivatives (returned in y2) of the spline function which is
  ! continuous in its second derivative and which obeys the boundary
  ! conditions.
  !
  ! In terms of the variable names used, the following equation is solved:
  !
  !		triDiag * y2 = b.
  !
  ! This routine was written with the help of Numerical Recipes for Fortran.
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: x(:), y(size(x)), loBoundValue, hiBoundValue
    character(*), intent(in) :: loBoundType, hiBoundType
    real, intent(out) :: y2(size(x))
    integer, intent(out), optional :: status

    integer :: localStatus, n, i
    real :: triDiag(-1:1, size(x)), b(size(x))

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    y2 = 0.0

    n = size(x)
  if (n <= 0) return ! no point in going further.

    select case(loBoundType)
    case('fixed1st') ! loBoundValue is taken to be the value of the
    ! first derivative at the lowest x sample.

      triDiag(0, 1) = (x(2) - x(1)) / 3.0
      triDiag(1, 1) = (x(2) - x(1)) / 6.0
      b(1) = (y(2) - y(1)) / (x(2) - x(1)) - loBoundValue
    case('fixed2nd') ! loBoundValue is taken to be the value of the
    ! second derivative at the lowest x sample.

      triDiag(0, 1) = 1.0
      triDiag(1, 1) = 0.0
      b(1) = loBoundValue

    case default
      if (present(status)) status = -1
  return
    end select

    select case(hiBoundType)
    case('fixed1st') ! hiBoundType is taken to be the value of the
    ! first derivative at the highest x sample.

      triDiag(-1, n) = (x(n) - x(n-1)) / 6.0
      triDiag( 0, n) = (x(n) - x(n-1)) / 3.0
      b(n) = hiBoundValue - (y(n) - y(n-1)) / (x(n) - x(n-1))

    case('fixed2nd') ! hiBoundType is taken to be the value of the
    ! second derivative at the highest x sample.

      triDiag(-1, n) = 0.0
      triDiag( 0, n) = 1.0
      b(n) = hiBoundValue

    case default
      if (present(status)) status = -1
  return
    end select

    ! Fill up the middle rows of triDiag and b:

    do i = 2, n - 1
      triDiag(-1, i) = (x( i ) - x(i-1)) / 6.0
      triDiag( 0, i) = (x(i+1) - x(i-1)) / 3.0
      triDiag( 1, i) = (x(i+1) - x( i )) / 6.0

      b(i) = (y(i+1) - y( i )) / (x(i+1) - x( i ))&
           - (y( i ) - y(i-1)) / (x( i ) - x(i-1))
    end do

    triDiag(-1, 1) = 0.0 ! always.
    triDiag( 1, n) = 0.0 ! always.

    ! Now solve the system of equations:

    call solveLinearTriDiag(triDiag, b, y2)

    if (present(status)) status = localStatus

  end subroutine calcCubicSplines

  !---------------------------------------------------------------------
  subroutine cubicSplineInterpolate(x, y, y2, xIn, yOut, status)
  !---------------------------------------------------------------------
  ! This is designed to work in conjunction with subroutine
  ! calcCubicSplines() to provide a cubic-spline-type interpolation.
  ! The present subroutine, given an ordered series of unique x values,
  ! plus the corresponding y values, plus the corresponding values of
  ! the second derivative y2, and finally plus a single x value xIn,
  ! performs the interpolation using the formula
  !
  !	y = a*y(i) + b*y(i+1) + c*y2(i) + d*y2(i+1)
  !
  ! where
  !
  !	     x(i+1) - xIn
  !	a = ---------------, b = 1 - a,
  !	     x(i+1) - x(i)
  !
  !	     1
  !	c = --- a*(a*a - 1)*(x(i+1) - x(i))^2
  !	     6
  ! and
  !	     1
  !	d = --- b*(b*b - 1)*(x(i+1) - x(i))^2
  !	     6
  !
  ! to return in the variable yOut the value of y at xIn.
  !
  ! Note that xIn must be >= x(1) and <= x(n).
  !
  ! This routine was written with the help of Numerical Recipes for Fortran.
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: x(:), y(size(x)), y2(size(x)), xIn
    real, intent(out) :: yOut
    integer, intent(out), optional :: status

    integer :: localStatus, n, i
    real :: a, b, denom, commonTerm

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    yOut = 0.0

    n = size(x)

    if (xIn < x(1) .or. xIn > x(n)) then
      if (present(status)) status = -1
  return
    end if

    i = largestIndex(x <= xIn) ! in array_utils
    if (i == n) i = n - 1

    denom = x(i+1) - x(i)
    a = (x(i+1) - xIn)  / denom
    b = (xIn    - x(i)) / denom

    commonTerm = denom * denom / 6.0
    yOut = a * (y(i)   + (a * a - 1.0) * commonTerm * y2(i))&
         + b * (y(i+1) + (b * b - 1.0) * commonTerm * y2(i+1))

    if (present(status)) status = localStatus

  end subroutine cubicSplineInterpolate

  !---------------------------------------------------------------------
  subroutine maskArrayLimits(mask, minXi, maxXi, minYi, maxYi)
  !---------------------------------------------------------------------
  ! Adapted from CGP's im_lims subroutine in implot.

    implicit none

    logical, intent(in) :: mask(:,:)
    integer, intent(out) :: minXi, maxXi, minYi, maxYi

    integer :: xsize, ysize, xi, yi

    !-------------------------------------------------------------------
    ! Start

    ! Defaults:
    minXi = 1
    maxXi = 0
    minYi = 1
    maxYi = 0
    
  if (count(mask) .le. 0) return

    xsize = size(mask, 1)
    ysize = size(mask, 2)

  if (xsize .le. 0 .or. ysize .le. 0) return

    do xi = 1, xsize
      if (any(mask(xi, :))) then
        minXi = xi
    exit
      end if
    end do

    do xi = xsize, 1, -1
      if (any(mask(xi, :))) then
        maxXi = xi
    exit
      end if
    end do

    do yi = 1, ysize
      if (any(mask(:, yi))) then
        minYi = yi
    exit
      end if
    end do

    do yi = ysize, 1, -1
      if (any(mask(:, yi))) then
        maxYi = yi
    exit
      end if
    end do

  end subroutine maskArrayLimits

  !---------------------------------------------------------------------
  subroutine maskArrayLimits_old(mask, minXi, maxXi, minYi, maxYi, status)
  !---------------------------------------------------------------------
  ! Adapted from CGP's im_lims subroutine in implot.

    implicit none

    logical, intent(in) :: mask(:,:)
    integer, intent(out) :: minXi, maxXi, minYi, maxYi
    integer, intent(out), optional :: status

    integer :: localStatus, xsize, ysize, xi, yi

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    minXi = 0
    maxXi = 0
    minYi = 0
    maxYi = 0
    
    if (count(mask) .le. 0) then
      if (present(status)) status = -1
  return
    end if

    xsize = size(mask, 1)
    ysize = size(mask, 2)

    if (xsize .le. 0 .or. ysize .le. 0) then
      if (present(status)) status = -2
  return
    end if

    do xi = 1, xsize
      if (any(mask(xi, :))) then
        minXi = xi
    exit
      end if
    end do
    do xi = xsize, 1, -1
      if (any(mask(xi, :))) then
        maxXi = xi
    exit
      end if
    end do
    do yi = 1, ysize
      if (any(mask(:, yi))) then
        minYi = yi
    exit
      end if
    end do
    do yi = ysize, 1, -1
      if (any(mask(:, yi))) then
        maxYi = yi
    exit
      end if
    end do

  end subroutine maskArrayLimits_old

  !---------------------------------------------------------------------
  function vectorCrossReal(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real, dimension(3), intent(in) :: vectorA, vectorB
    real :: vectorCrossReal(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossReal(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossReal(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossReal(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossReal

  !---------------------------------------------------------------------
  function vectorCrossDble(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), dimension(3), intent(in) :: vectorA, vectorB
    real(kind(1d0)) :: vectorCrossDble(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossDble(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossDble(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossDble(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossDble

  !---------------------------------------------------------------------
  function trapezoidSum(x, y) result(summ)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: x(:), y(size(x))
    real :: summ

    integer :: i, n

    !-------------------------------------------------------------------
    ! Start

    summ = 0.0 ! default
    n = size(x)
  if (n <= 1) return

    do i = 1, n - 1
      summ = summ + x(i + 1) * y(i) - x(i) * y(i + 1)
    end do
    summ = 0.5 * (x(n) * y(n) - x(1) * y(1) + summ)

  end function trapezoidSum

  !---------------------------------------------------------------------
  function getBinNum(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------
  ! At present this thing does as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if this is necessary.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if this is
  ! necessary.
  !.....................................................................

    implicit none

    real, intent(in) :: xBinEdges(:), x
    integer :: binNum

    !-------------------------------------------------------------------
    ! Start

    if (size(xBinEdges) < 2) then
      binNum = 0
  return
    end if

    binNum = largestIndex(xBinEdges <= x)
!print *, 'x, binNum: ', x, binNum
!    if (binNum == size(xBinEdges)) binNum = binNum - 1

  end function getBinNum

  !---------------------------------------------------------------------
  integer function smallestIndex(maskVector)
  !---------------------------------------------------------------------
!*** note I have swapped around the default values of smallestIndex and largestIndex.

    implicit none

    logical, intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    smallestIndex = size(maskVector) + 1
    do i = 1, size(maskVector)
    if (maskVector(i)) then
        smallestIndex = i
    exit
      end if
    end do

  end function smallestIndex

  !---------------------------------------------------------------------
  integer function largestIndex(maskVector)
  !---------------------------------------------------------------------
!*** note I have swapped around the default values of smallestIndex and largestIndex. Now, if NONE of the elements of mask() are true, 0 is returned. NOTE that the behaviour of getBinNum() depends on this, so don't change the default back!

    implicit none

    logical, intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    largestIndex = 0
    do i = size(maskVector), 1, -1
      if (maskVector(i)) then
        largestIndex = i
    exit
      end if
    end do

  end function largestIndex

  !---------------------------------------------------------------------
  subroutine rebinImageBasicReal(oldImage, newImage, status)
  !---------------------------------------------------------------------
  ! The rebinning is done in two passes, x first then y, using the math_utils
  ! procedure rebin (cut and pasted in here until math_utils is accepted).

    implicit none

    real, intent(in) ::&
      oldImage(:,:)
    real, intent(out) ::&
      newImage(:,:)
    integer, intent(out), optional :: status

    real ::&
      tempImage(size(oldImage, 1), size(newImage, 2))
    integer :: localStatus, foreignStatus, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    tempImage = 0.0 ! array op, default
    do xi = 1, size(oldImage, 1)
      call rebin(oldImage(xi, :), tempImage(xi, :), status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 0)
!***
      end if
    end do

    newImage = 0.0 ! array op, default
    do yi = 1, size(newImage, 2)
      call rebin(tempImage(:, yi), newImage(:, yi), status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 1)
!***
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine rebinImageBasicReal

  !---------------------------------------------------------------------
  subroutine rebinImageBasicDble(oldImage, newImage, status)
  !---------------------------------------------------------------------
  ! The rebinning is done in two passes, x first then y, using the math_utils
  ! procedure rebin (cut and pasted in here until math_utils is accepted).

    implicit none

    real(kind(1d0)), intent(in) ::&
      oldImage(:,:)
    real(kind(1d0)), intent(out) ::&
      newImage(:,:)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      tempImage(size(oldImage, 1), size(newImage, 2))
    integer :: localStatus, foreignStatus, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    tempImage = 0.0 ! array op, default
    do xi = 1, size(oldImage, 1)
      call rebin(oldImage(xi, :), tempImage(xi, :), status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 0)
!***
      end if
    end do

    newImage = 0.0 ! array op, default
    do yi = 1, size(newImage, 2)
      call rebin(tempImage(:, yi), newImage(:, yi), status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 1)
!***
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine rebinImageBasicDble

  !---------------------------------------------------------------------
  subroutine rebinImageWithEdgesReal(oldImage, oldXedges, oldYedges&
  , newImage, newXedges, newYedges, status)
  !---------------------------------------------------------------------
  ! The rebinning is done in two passes, x first then y, using the math_utils
  ! procedure rebin (cut and pasted in here until math_utils is accepted).

    implicit none

    real, intent(in) ::&
      oldImage(:,:),&
      oldXedges(size(oldImage, 1) + 1), oldYedges(size(oldImage, 2) + 1),&
      newXedges(:), newYedges(:)
    real, intent(out) ::&
      newImage(size(newXedges) - 1, size(newYedges) - 1)
    integer, intent(out), optional :: status

    real ::&
      tempImage(size(oldImage, 1), size(newImage, 2))
    integer :: localStatus, foreignStatus, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    tempImage = 0.0 ! array op, default
    do xi = 1, size(oldImage, 1)
      call rebin(oldImage(xi, :), tempImage(xi, :), oldYedges, newYedges&
        , foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 0)
!***
      end if
    end do

    newImage = 0.0 ! array op, default
    do yi = 1, size(newImage, 2)
      call rebin(tempImage(:, yi), newImage(:, yi), oldXedges, newXedges&
        , status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 1)
!***
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine rebinImageWithEdgesReal

  !---------------------------------------------------------------------
  subroutine rebinImageWithEdgesDble(oldImage, oldXedges, oldYedges&
  , newImage, newXedges, newYedges, status)
  !---------------------------------------------------------------------
  ! The rebinning is done in two passes, x first then y, using the math_utils
  ! procedure rebin (cut and pasted in here until math_utils is accepted).

    implicit none

    real(kind(1d0)), intent(in) ::&
      oldImage(:,:),&
      oldXedges(size(oldImage, 1) + 1), oldYedges(size(oldImage, 2) + 1),&
      newXedges(:), newYedges(:)
    real(kind(1d0)), intent(out) ::&
      newImage(size(newXedges) - 1, size(newYedges) - 1)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      tempImage(size(oldImage, 1), size(newImage, 2))
    integer :: localStatus, foreignStatus, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    tempImage = 0.0 ! array op, default
    do xi = 1, size(oldImage, 1)
      call rebin(oldImage(xi, :), tempImage(xi, :), oldYedges, newYedges&
        , foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 0)
!***
      end if
    end do

    newImage = 0.0 ! array op, default
    do yi = 1, size(newImage, 2)
      call rebin(tempImage(:, yi), newImage(:, yi), oldXedges, newXedges&
        , status=foreignStatus)

      if (foreignStatus .lt. 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus .gt. 0) then
        localStatus = ibset(localStatus, 1)
!***
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine rebinImageWithEdgesDble

  !---------------------------------------------------------------------
  subroutine rebinBasicReal(oldYvalues, newYvalues, status)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in) ::&
      oldYvalues(:)
    real, intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, newBin, localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    do oldBin = 1, numOldBins + 1
      oldXedges(oldBin) = (oldBin - 1) / real(numOldBins)
    end do
    do newBin = 1, numNewBins + 1
      newXedges(newBin) = (newBin - 1) / real(numNewBins)
    end do

    call rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
    , localStatus)

    if (present(status)) status = localStatus

  end subroutine rebinBasicReal

  !---------------------------------------------------------------------
  subroutine rebinBasicDble(oldYvalues, newYvalues, status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldYvalues(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, newBin, localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    do oldBin = 1, numOldBins + 1
      oldXedges(oldBin) = (oldBin - 1) / real(numOldBins)
    end do
    do newBin = 1, numNewBins + 1
      newXedges(newBin) = (newBin - 1) / real(numNewBins)
    end do

    call rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
    , localStatus)

    if (present(status)) status = localStatus

  end subroutine rebinBasicDble

  !---------------------------------------------------------------------
  subroutine rebinWithWidthsReal(oldYvalues, newYvalues, oldXwidths&
  , newXwidths, oldXcentres, newXcentres, status)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in), dimension(:) ::&
      oldXwidths, oldXcentres, oldYvalues, newXwidths, newXcentres
    real, intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins,&
      localStatus, foreignStatus
    integer ::&
      oldBin, newBin

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXwidths) .lt. numOldBins) then
      if(present(status)) status = -1
  return
    else if (size(oldXwidths) .gt. numOldBins) then
      localStatus = localStatus + 1
    end if
    if (size(newXwidths) .lt. numNewBins) then
      if(present(status)) status = -2
  return
    else if (size(newXwidths) .gt. numNewBins) then
      localStatus = localStatus + 2
    end if

    if (size(oldXcentres) .lt. numOldBins) then
      if(present(status)) status = -3
  return
    else if (size(oldXcentres) .gt. numOldBins) then
      localStatus = localStatus + 4
    end if
    if (size(newXcentres) .lt. numNewBins) then
      if(present(status)) status = -4
  return
    else if (size(newXcentres) .gt. numNewBins) then
      localStatus = localStatus + 8
    end if

    do oldBin = 1, numOldBins
      oldXedges(oldBin) = oldXcentres(oldBin) - 0.5 * oldXwidths(oldBin)
    end do
    oldXedges(numOldBins + 1) = oldXcentres(numOldBins) + 0.5&
    * oldXwidths(numOldBins)

    do newBin = 1, numNewBins
      newXedges(newBin) = newXcentres(newBin) - 0.5 * newXwidths(newBin)
    end do
    newXedges(numNewBins + 1) = newXcentres(numNewBins) + 0.5&
    * newXwidths(numNewBins)

    call rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
    , foreignStatus)
    if (foreignStatus .lt. 0) then
      if(present(status)) status = foreignStatus - 4
  return
    end if

    localStatus = localStatus + foreignStatus * 16
    if(present(status)) status = localStatus

  end subroutine rebinWithWidthsReal

  !---------------------------------------------------------------------
  subroutine rebinWithWidthsDble(oldYvalues, newYvalues, oldXwidths&
  , newXwidths, oldXcentres, newXcentres, status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(1d0)), intent(in), dimension(:) ::&
      oldXwidths, oldXcentres, oldYvalues, newXwidths, newXcentres
    real(kind(1d0)), intent(out) ::&
      newYvalues(:)
    integer, intent(out), optional :: status

    real(kind(1d0)) ::&
      oldXedges(size(oldYvalues) + 1), newXedges(size(newYvalues) + 1)
    integer ::&
      numOldBins,&
      numNewBins,&
      localStatus, foreignStatus
    integer ::&
      oldBin, newBin

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXwidths) .lt. numOldBins) then
      if(present(status)) status = -1
  return
    else if (size(oldXwidths) .gt. numOldBins) then
      localStatus = localStatus + 1
    end if
    if (size(newXwidths) .lt. numNewBins) then
      if(present(status)) status = -2
  return
    else if (size(newXwidths) .gt. numNewBins) then
      localStatus = localStatus + 2
    end if

    if (size(oldXcentres) .lt. numOldBins) then
      if(present(status)) status = -3
  return
    else if (size(oldXcentres) .gt. numOldBins) then
      localStatus = localStatus + 4
    end if
    if (size(newXcentres) .lt. numNewBins) then
      if(present(status)) status = -4
  return
    else if (size(newXcentres) .gt. numNewBins) then
      localStatus = localStatus + 8
    end if

    do oldBin = 1, numOldBins
      oldXedges(oldBin) = oldXcentres(oldBin) - 0.5d0 * oldXwidths(oldBin)
    end do
    oldXedges(numOldBins + 1) = oldXcentres(numOldBins) + 0.5d0&
    * oldXwidths(numOldBins)

    do newBin = 1, numNewBins
      newXedges(newBin) = newXcentres(newBin) - 0.5d0 * newXwidths(newBin)
    end do
    newXedges(numNewBins + 1) = newXcentres(numNewBins) + 0.5d0&
    * newXwidths(numNewBins)

    call rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
    , foreignStatus)
    if (foreignStatus .lt. 0) then
      if(present(status)) status = foreignStatus - 4
  return
    end if

    localStatus = localStatus + foreignStatus * 16
    if(present(status)) status = localStatus

  end subroutine rebinWithWidthsDble

  !---------------------------------------------------------------------
  subroutine rebinWithEdgesReal(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------

    implicit none 

    real, intent(in) ::&
      oldXedges(:), oldYvalues(size(oldXedges)-1), newXedges(:)
    real, intent(out) ::&
      newYvalues(size(newXedges)-1)
    integer, intent(out), optional :: status

    integer :: foreignStatus
    real(kind(1d0)) :: newYvaluesDble(size(newYvalues))

    !-------------------------------------------------------------------
    ! Start:

    call rebinKernel(dble(oldYvalues), newYvaluesDble, dble(oldXedges)&
    , dble(newXedges), foreignStatus)

    if (present(status)) status = foreignStatus
    newYvalues = real(newYvaluesDble)

  end subroutine rebinWithEdgesReal

  !---------------------------------------------------------------------
  subroutine rebinWithEdgesDble(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldXedges(:), oldYvalues(size(oldXedges)-1), newXedges(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(size(newXedges)-1)
    integer, intent(out), optional :: status

    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start:

    call rebinKernel(oldYvalues, newYvalues, oldXedges&
    , newXedges, foreignStatus)

    if (present(status)) status = foreignStatus

  end subroutine rebinWithEdgesDble

  !---------------------------------------------------------------------
  subroutine rebinKernel(oldYvalues, newYvalues, oldXedges, newXedges&
  , status)
  !---------------------------------------------------------------------
  ! This is designed to be a pretty general and robust rebinning routine.
  ! Neither the old or new series is constrained to have bins of equal width.

    implicit none 

    real(kind(1d0)), intent(in) ::&
      oldXedges(:), oldYvalues(size(oldXedges)-1), newXedges(:)
    real(kind(1d0)), intent(out) ::&
      newYvalues(size(newXedges)-1)
    integer, intent(out) :: status

    integer ::&
      numOldBins,&
      numNewBins
    integer ::&
      oldBin, oldStartBin, oldFinisBin, newBin
    logical ::&
      foundOverlappingBin
    real(kind(1d0)) ::&
      overlap, binWidth

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default

    numOldBins = size(oldYvalues)
    numNewBins = size(newYvalues)

    if (size(oldXedges) .lt. numOldBins + 1) then
      status = -1
  return
    else if (size(oldXedges) .gt. numOldBins + 1) then
      status = status + 1
    end if
    if (size(newXedges) .lt. numNewBins + 1) then
      status = -2
  return
    else if (size(newXedges) .gt. numNewBins + 1) then
      status = status + 2
    end if

    do oldBin = 1, numOldBins
      binWidth = oldXedges(oldBin + 1) - oldXedges(oldBin)
      if (binWidth .le. 0.0) then
!print *, 'oldBin:', oldBin
!print *, 'oldXedges(oldBin + 1):', oldXedges(oldBin + 1)
!print *, 'oldXedges(oldBin):', oldXedges(oldBin)
        status = -3
  return
      end if
    end do

    do newBin = 1, numNewBins
      binWidth = newXedges(newBin + 1) - newXedges(newBin)
      if (binWidth .le. 0.0) then
        status = -4
  return
      end if
    end do

    newYvalues = 0.0 ! vector op, default.

    oldStartBin = 1
    NEW: do newBin = 1, numNewBins
      ! Find first old bin that overlaps newBin:
      foundOverlappingBin = .false. ! default
      OLD1: do oldBin = oldStartBin, numOldBins
        if (oldXedges(oldBin + 1) .gt. newXedges(newBin)&
        .and. oldXedges(oldBin) .lt. newXedges(newBin + 1)) then
          foundOverlappingBin = .true.
          oldStartBin = oldBin
      exit OLD1
        end if
      end do OLD1

      if (.not. foundOverlappingBin) then ! have either reached the end of the
      ! old series, or not reached the start. Leave new values at default 0.0.
    exit NEW
      end if

      ! Find the last old bin that overlaps newBin:
      foundOverlappingBin = .true. ! default
      oldFinisBin = numOldBins ! default
      OLD2: do oldBin = oldStartBin + 1, numOldBins
        if (oldXedges(oldBin) .ge. newXedges(newBin + 1)) then
          foundOverlappingBin = .false.
          oldFinisBin = oldBin - 1
      exit OLD2
        end if
      end do OLD2

      if (foundOverlappingBin) then ! have reached the end of the
      ! old series without leaving the new bin.
        oldFinisBin = numOldBins
      end if

      ! Now work though all the old bins that overlap the new, and pour in
      ! the correct proportional amounts from them into newBin:
      OLD3: do oldBin = oldStartBin, oldFinisBin
        if (oldXedges(oldBin + 1) .lt. newXedges(newBin + 1)) then
          if (oldXedges(oldBin) .gt. newXedges(newBin)) then !B - all
          ! of old bin goes into new:
            overlap = oldXedges(oldBin + 1) - oldXedges(oldBin)

          else !A - upper part of old goes into new:
            overlap = oldXedges(oldBin + 1) - newXedges(newBin)
          end if

        else
          if (oldXedges(oldBin) .gt. newXedges(newBin)) then !D - lower
          ! part of old goes into new:
            overlap = newXedges(newBin + 1) - oldXedges(oldBin)

          else !C - middle part of old goes into new:
            overlap = newXedges(newBin + 1) - newXedges(newBin)
          end if
        end if
        newYvalues(newBin) = newYvalues(newBin) + oldYvalues(oldBin)&
        * overlap
      end do OLD3

      oldStartBin = oldFinisBin
    end do NEW

    ! Divide by the new bin width:
    do newBin = 1, numNewBins
      binWidth = newXedges(newBin + 1) - newXedges(newBin)
      newYvalues(newBin) = newYvalues(newBin) / binWidth
    end do

  end subroutine rebinKernel

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module array_utils
