! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: sensitivity
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module sensitivity

  use dal
!  use param  
!  use errorHandling

  use math_utils

  implicit none 

  private 

!  public minExpectedPoissonCounts
  public minDetPoissonCounts

  interface minDetPoissonCounts
    module procedure minDetPoissonCountsScalar, minDetPoissonCountsVector
  end interface

  integer, public :: numBands = 0 ! default value.
  real, allocatable, public :: globalSrcCountRatios(:), globalBkgCounts(:)
  real, public :: targetLikelihood = 0.0, targetProb = 0.0 ! default values.
  real, public :: globalBkgCount = 0.0 ! default value.

contains 

  !---------------------------------------------------------------------
  subroutine minDetPoissonCountsScalar(bkgCount, likelihoodCutoff&
  , detectableCounts, detectableCountsUncert, status)
  !---------------------------------------------------------------------
  ! Given a probability distribution P of event counts having an average value a, any sample value c is associated with a certain probability P_int (and therefore likelihood L = -ln(P_int)) of it _not_ being due to chance. This probability is obtained by summing or integrating the distribution from j=c to j=infinity, which sum happens, for the Poisson distribution assumed here, to be equal to the incomplete Gamma function P(c,a). (Note that this is valid whether c is integer or not.)
  !
  ! The present subroutine inverts this relation to return that value of c (=counts) which is associated with the specified pair (L,a). (Ideally there would be some nifty series for this inversion but I haven't come across one. Failing that I would have liked to perform this inversion by use of Newton's method but I can't find an expression for dP_int/dc. Ridders' method is the best I can come up with at present.)
  !
  ! Notes:
  !	(1) The value detectableCounts which is returned is NOT pure source counts, but source + background.
!*** this differs from minDetPoissonCountsVector!!!
  !	(2) This routine is ONLY valid if the background can itself be assumed to be a known value (ie, of insignificant uncertainty). If the background has had to be estimated then the stats get more complicated and simple Poisson statistics no longer rule the waves.
  !

    implicit none

    real,    intent(in)            :: bkgCount, likelihoodCutoff
    real,    intent(out)           :: detectableCounts
    real,    intent(out), optional :: detectableCountsUncert
    integer, intent(out), optional :: status

    real :: prob, oldCounts, counts, counts0, counts1, convergenceCriterion
    integer :: localStatus, foreignStatus

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    detectableCounts       = 0.0 ! default
    detectableCountsUncert = 0.0 ! default

    ! Check for wrong and special arguments:

    if (likelihoodCutoff <= 0.0) then
      if (present(status)) status = -1
  return
    end if
    if (bkgCount <= 0.0) then
      if (present(status)) status = -2
  return
    end if

    !...................................................................
    ! Set up globals:

    targetLikelihood = likelihoodCutoff
    targetProb = exp(-likelihoodCutoff)
    globalBkgCount = bkgCount

    !...................................................................
    ! Start with counts = 1.0 and multiply or divide it by 2 until the
    ! solution is bracketted:

    oldCounts = 1.0
    call incompleteGammaP(oldCounts, bkgCount, prob)
    if (prob == targetProb) then
      detectableCounts = oldCounts
  return

    else if (prob < targetProb) then ! decrease the counts:
      do
        counts = oldCounts / 2.0
        call incompleteGammaP(counts, bkgCount, prob)
      if (prob >= targetProb) exit
        oldCounts = counts
      end do

      if (prob == targetProb) then
        detectableCounts = counts
  return
      end if

      counts0 = counts
      counts1 = oldCounts

    else ! increase the counts:
      do
        counts = oldCounts * 2.0
        call incompleteGammaP(counts, bkgCount, prob)
      if (prob <= targetProb) exit
        oldCounts = counts
      end do

      if (prob == targetProb) then
        detectableCounts = counts
  return
      end if

      counts0 = counts
      counts1 = oldCounts
    end if

    !...................................................................
    ! Do Ridders'-method inversion:

    convergenceCriterion = (counts0 + counts1) / 2000.0

    call riddersMethod(counts0, counts1, yScalar, detectableCounts&
      , maxNumIterations=30, convergenceCriterion=convergenceCriterion&
      , dx=detectableCountsUncert, status=foreignStatus)

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus - 2
      detectableCounts = 0.0
  return
    else if (foreignStatus > 0) then
      localStatus = foreignStatus
    end if

    if (present(status)) status = localStatus

  end subroutine minDetPoissonCountsScalar

  !---------------------------------------------------------------------
  real function yScalar(a)
  !---------------------------------------------------------------------
  ! Note that globalBkgCount and targetLikelihood are globals and should be given sensible values before the function is called. There is no protection in this function for undefined or incorrect values.

    implicit none

    real, intent(in) :: a

    real :: prob

    !-------------------------------------------------------------------
    ! Start

    call incompleteGammaP(a, globalBkgCount, prob)
    yScalar = targetLikelihood + log(prob)

  end function yScalar

  !---------------------------------------------------------------------
  subroutine minDetPoissonCountsVector(srcCountRatios, bkgCounts&
  , likelihoodCutoff, detectableCounts, detectableCountsUncert, status)
  !---------------------------------------------------------------------
  ! Given a probability distribution P of event counts having an average value a, any sample value c is associated with a certain probability P_int (and therefore likelihood L = -ln(P_int)) of it _not_ being due to chance. This probability is obtained by summing or integrating the distribution from j=c to j=infinity, which sum happens, for the Poisson distribution assumed here, to be equal to the incomplete Gamma function P(c,a). (Note that this is valid whether c is integer or not.)
  !
  ! The present subroutine inverts this relation to return that value of c (=counts) which is associated with the specified pair (L,a). (Ideally there would be some nifty series for this inversion but I haven't come across one. Failing that I would have liked to perform this inversion by use of Newton's method but I can't find an expression for dP_int/dc. Ridders' method is the best I can come up with at present.)
  !
  ! Notes:
  !	(1) The value detectableCounts which is returned is the pure source counts, ie total counts minus background.
  !	(2) This routine is ONLY valid if the background can itself be assumed to be a known value (ie, of insignificant uncertainty). If the background has had to be estimated then the stats get more complicated and simple Poisson statistics no longer rule the waves.
  !	(3) The subroutine cannot be called twice from the same program with a different value of numBands unless this is set to zero and the vectors globalSrcCountRatios and globalBkgCounts deallocated in between calls.
  !

    implicit none

    real,    intent(in)            :: srcCountRatios(:),&
                                      bkgCounts(size(srcCountRatios)),&
                                      likelihoodCutoff
    real,    intent(out)           :: detectableCounts(size(srcCountRatios))
    real,    intent(out), optional ::&
      detectableCountsUncert(size(srcCountRatios))
    integer, intent(out), optional :: status

    real ::&
      prob, totalProb,&
      oldCounts, counts, counts0, counts1, countsUncert,&
      convergenceCriterion
    integer :: band, localStatus, foreignStatus
    logical :: firstTime = .true. ! preset variables are saved

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    if (present(status)) status = localStatus
    detectableCounts       = 0.0 ! default
    detectableCountsUncert = 0.0 ! default

    ! Check for wrong and special arguments:

    if (likelihoodCutoff <= 0.0) then
      if (present(status)) status = -1
  return
    end if
    if (size(srcCountRatios) <= 0) then
      if (present(status)) status = -2
  return
    end if
    if (any(srcCountRatios <= 0.0)) then
      if (present(status)) status = -3
  return
    end if
    if (any(bkgCounts <= 0.0)) then
      if (present(status)) status = -4
  return
    end if

    ! Check that >0 source counts returns a value of likelihood which
    ! is greater than the target one.

    totalProb = 1.0
    do band = 1, numBands
      call incompleteGammaP(bkgCounts(band), bkgCounts(band), prob)
      totalProb = totalProb * prob
!if (totalProb < targetProb) then
!print *, 'band, bkgCounts(band), prob, totalProb, targetProb:'&
!, band, bkgCounts(band), prob, totalProb, targetProb
!end if
    end do
    if (totalProb < targetProb) then
      if (present(status)) status = -5
  return
    end if

    !...................................................................
    ! Set up globals:

    targetLikelihood = likelihoodCutoff
    targetProb = exp(-likelihoodCutoff)
    numBands = size(srcCountRatios)
    if (firstTime) then ! then globalSrcCountRatios and globalBkgCounts
    ! have not yet been allocated. Do so:
      firstTime = .false.
      allocate(globalSrcCountRatios(numBands), globalBkgCounts(numBands))
    end if
    globalSrcCountRatios = srcCountRatios
    globalBkgCounts = bkgCounts

    !...................................................................
    ! Start with counts = 1.0 and multiply or divide it by 2 until the
    ! solution is bracketted:

    oldCounts = 1.0
    totalProb = 1.0
    do band = 1, numBands
      call incompleteGammaP(oldCounts * srcCountRatios(band) + bkgCounts(band)&
        , bkgCounts(band), prob)
      totalProb = totalProb * prob
    end do
    if (totalProb == targetProb) then
      detectableCounts = oldCounts * srcCountRatios
  return

    else if (totalProb < targetProb) then ! decrease the counts:
      do
        counts = oldCounts / 2.0
        totalProb = 1.0
        do band = 1, numBands
          call incompleteGammaP(counts * srcCountRatios(band) + bkgCounts(band)&
            , bkgCounts(band), prob)
          totalProb = totalProb * prob
        end do
      if (totalProb >= targetProb) exit
        oldCounts = counts
      end do

      if (totalProb == targetProb) then
        detectableCounts = counts * srcCountRatios
  return
      end if

      counts0 = counts
      counts1 = oldCounts

    else ! increase the counts:
      do
        counts = oldCounts * 2.0
        totalProb = 1.0
        do band = 1, numBands
          call incompleteGammaP(counts * srcCountRatios(band) + bkgCounts(band)&
            , bkgCounts(band), prob)
          totalProb = totalProb * prob
        end do
      if (totalProb <= targetProb) exit
        oldCounts = counts
      end do

      if (totalProb == targetProb) then
        detectableCounts = counts * srcCountRatios
  return
      end if

      counts0 = counts
      counts1 = oldCounts
    end if

    !...................................................................
    ! Do Ridders'-method inversion:

    convergenceCriterion = (counts0 + counts1) / 2000.0

    call riddersMethod(counts0, counts1, yVector, counts&
      , maxNumIterations=30, convergenceCriterion=convergenceCriterion&
      , dx=countsUncert, status=foreignStatus)

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus - 5
      detectableCounts = 0.0
  return
    else if (foreignStatus > 0) then
      localStatus = foreignStatus
    end if

    detectableCounts       = counts       * srcCountRatios
    detectableCountsUncert = countsUncert * srcCountRatios

    if (present(status)) status = localStatus

  end subroutine minDetPoissonCountsVector

  !---------------------------------------------------------------------
  real function yVector(a)
  !---------------------------------------------------------------------
  ! Note that globalBkgCounts, globalSrcCountRatios and targetLikelihood are globals and should be given sensible values before the function is called. There is no protection in this function for undefined or incorrect values.

    implicit none

    real, intent(in) :: a

    real :: totalProb, prob
    integer :: band

    !-------------------------------------------------------------------
    ! Start

    totalProb = 1.0
    do band = 1, numBands
      call incompleteGammaP(a * globalSrcCountRatios(band)&
        + globalBkgCounts(band), globalBkgCounts(band), prob)
      totalProb = totalProb * prob
    end do
    yVector = targetLikelihood + log(totalProb)

  end function yVector

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module sensitivity
