! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: wcs_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Returns semi-axes of an ellipse round each of a list of sources.
! 
! Comments: TODO:
! 
! $Id: wcs_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module wcs_aux

  use dal
  use errorHandling ! not needed now but should replace 'noisy' messages.

  use constants
!  use coordinate ! for tanToRaDec.
  use fits_utils ! contains subs getRealKwd(), getRealKwd(), getStringKwd().

  implicit none 

  private 

  public getWcsInfo
  public easifyWcsType
  public wcsAreSame
  public dumpWcs
  public tanToPixels
  public tanToPixelsAxis
  public pixelsToTan

!public getWcsInfoSourceListSet

  interface getWcsInfo
    module procedure getWcsInfo_image, getWcsInfoArray
    module procedure getWcsInfoSet, getWcsInfoSetName, getWcsInfoSetArrayName
    module procedure getWcsInfo_table, getWcsInfoSetArrayNumber
  end interface

  interface easifyWcsType
    module procedure easifyWcsType_scalar
    module procedure easifyWcsType_vector
  end interface

  interface tanToPixels
    module procedure tanToPixelsScalar, tanToPixelsVector
  end interface

  interface tanToPixelsAxis
    module procedure tanToPixelsAxisScalar, tanToPixelsAxisVector
    module procedure tanToPixelsAxisArray
  end interface

  interface pixelsToTan
    module procedure pixelsToTanScalar, pixelsToTanVector
  end interface

  type, public :: wcsAxisType
!    integer(int32) ::&
!      refPixels
    character(len=8) ::&
      ctype
    real(double) ::&
      refPixels,&
      pixelSizeWorld,&
      refWorld
    integer(int8) ::&
      pixelSizeSign
  end type wcsAxisType

  type, public :: wcsDataType
    type(wcsAxisType) :: x, y
  end type wcsDataType

contains 

  !---------------------------------------------------------------------
  subroutine dumpWcs(wcs, verbosity)
  !---------------------------------------------------------------------

    implicit none

    type(wcsDataType), intent(in) :: wcs
    integer(int32), intent(in) :: verbosity

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, verbosity, 'WCS dump:')
    call message(AppMsg, verbosity, '  X axis:')
    call message(AppMsg, verbosity, '    CTYPE: '//wcs%x%ctype)
    write(errstr, '(a, es23.16)') '    CRVAL:', wcs%x%refWorld
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)') '    CRPIX:', wcs%x%refPixels
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)') '    CDELT:', wcs%x%pixelSizeWorld
    call message(AppMsg, verbosity)

    call message(AppMsg, verbosity, '  Y axis:')
    call message(AppMsg, verbosity, '    CTYPE: '//wcs%y%ctype)
    write(errstr, '(a, es23.16)') '    CRVAL:', wcs%y%refWorld
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)') '    CRPIX:', wcs%y%refPixels
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)') '    CDELT:', wcs%y%pixelSizeWorld
    call message(AppMsg, verbosity)
!print *, wcs%y%pixelSizeWorld

  end subroutine dumpWcs

  !---------------------------------------------------------------------
  logical function wcsAreSame(wcs1, wcs2)
  !---------------------------------------------------------------------

    implicit none

    type(wcsDataType), intent(in) :: wcs1, wcs2

    !-------------------------------------------------------------------
    ! Start

    wcsAreSame = .false. ! default

  if (wcs1%x%ctype          .ne. wcs2%x%ctype         ) return
  if (wcs1%x%refPixels      .ne. wcs2%x%refPixels     ) return
  if (wcs1%x%pixelSizeWorld .ne. wcs2%x%pixelSizeWorld) return
  if (wcs1%x%refWorld       .ne. wcs2%x%refWorld      ) return

  if (wcs1%y%ctype          .ne. wcs2%y%ctype         ) return
  if (wcs1%y%refPixels      .ne. wcs2%y%refPixels     ) return
  if (wcs1%y%pixelSizeWorld .ne. wcs2%y%pixelSizeWorld) return
  if (wcs1%y%refWorld       .ne. wcs2%y%refWorld      ) return

    wcsAreSame = .true.

  end function wcsAreSame

  !---------------------------------------------------------------------
  subroutine getWcsInfoSetName(setName, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*),      intent(in)  :: setName
    type(wcsDataType), intent(out) :: wcs

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, 0, READ)

    call getRealKwd("CRPIX1",   set, inArray, wcs%X%refPixels)
    call getRealKwd("CRPIX2",   set, inArray, wcs%Y%refPixels)
    call getRealKwd("CDELT1",   set, inArray, wcs%X%pixelSizeWorld)
    call getRealKwd("CDELT2",   set, inArray, wcs%Y%pixelSizeWorld)
    call getRealKwd("CRVAL1",   set, inArray, wcs%X%refWorld)
    call getRealKwd("CRVAL2",   set, inArray, wcs%Y%refWorld)
    call getStringKwd("CTYPE1", set, inArray, wcs%X%ctype)
    call getStringKwd("CTYPE2", set, inArray, wcs%Y%ctype)

    call release(set)

  end subroutine getWcsInfoSetName

  !---------------------------------------------------------------------
  subroutine getWcsInfoSetArrayName(setName, arrayName, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*),      intent(in)  :: setName, arrayName
    type(wcsDataType), intent(out) :: wcs

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, arrayName, READ)

    call getRealKwd("CRPIX1",   set, inArray, wcs%X%refPixels)
    call getRealKwd("CRPIX2",   set, inArray, wcs%Y%refPixels)
    call getRealKwd("CDELT1",   set, inArray, wcs%X%pixelSizeWorld)
    call getRealKwd("CDELT2",   set, inArray, wcs%Y%pixelSizeWorld)
    call getRealKwd("CRVAL1",   set, inArray, wcs%X%refWorld)
    call getRealKwd("CRVAL2",   set, inArray, wcs%Y%refWorld)
    call getStringKwd("CTYPE1", set, inArray, wcs%X%ctype)
    call getStringKwd("CTYPE2", set, inArray, wcs%Y%ctype)

    call release(set)

  end subroutine getWcsInfoSetArrayName

  !---------------------------------------------------------------------
  subroutine getWcsInfoSetArrayNumber(setName, arrayNumber, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*),      intent(in)  :: setName
    integer(int32),    intent(in)  :: arrayNumber
    type(wcsDataType), intent(out) :: wcs

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, arrayNumber, READ)

    call getRealKwd("CRPIX1",   set, inArray, wcs%X%refPixels)
    call getRealKwd("CRPIX2",   set, inArray, wcs%Y%refPixels)
    call getRealKwd("CDELT1",   set, inArray, wcs%X%pixelSizeWorld)
    call getRealKwd("CDELT2",   set, inArray, wcs%Y%pixelSizeWorld)
    call getRealKwd("CRVAL1",   set, inArray, wcs%X%refWorld)
    call getRealKwd("CRVAL2",   set, inArray, wcs%Y%refWorld)
    call getStringKwd("CTYPE1", set, inArray, wcs%X%ctype)
    call getStringKwd("CTYPE2", set, inArray, wcs%Y%ctype)

    call release(set)

  end subroutine getWcsInfoSetArrayNumber

  !---------------------------------------------------------------------
  subroutine getWcsInfo_image(set, inArray, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    type(DataSetT), intent(in) :: set
    type(ArrayT),   intent(in) :: inArray
    type(wcsDataType), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    call getRealKwd("CRPIX1",   set, inArray, wcs%X%refPixels)
    call getRealKwd("CRPIX2",   set, inArray, wcs%Y%refPixels)
    call getRealKwd("CDELT1",   set, inArray, wcs%X%pixelSizeWorld)
    call getRealKwd("CDELT2",   set, inArray, wcs%Y%pixelSizeWorld)
    call getRealKwd("CRVAL1",   set, inArray, wcs%X%refWorld)
    call getRealKwd("CRVAL2",   set, inArray, wcs%Y%refWorld)
    call getStringKwd("CTYPE1", set, inArray, wcs%X%ctype)
    call getStringKwd("CTYPE2", set, inArray, wcs%Y%ctype)

  end subroutine getWcsInfo_image

  !---------------------------------------------------------------------
  subroutine getWcsInfoArray(inArray, wcs, status)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT),   intent(in) :: inArray
    type(wcsDataType), intent(out) :: wcs
    integer,           intent(out), optional :: status

    integer :: localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    if (.not. (hasAttribute(inArray, 'CRPIX1')&
         .and. hasAttribute(inArray, 'CRPIX2')&
         .and. hasAttribute(inArray, 'CDELT1')&
         .and. hasAttribute(inArray, 'CDELT2')&
         .and. hasAttribute(inArray, 'CRVAL1')&
         .and. hasAttribute(inArray, 'CRVAL2')&
         .and. hasAttribute(inArray, 'CTYPE1')&
         .and. hasAttribute(inArray, 'CTYPE2'))) then
      if(present(status)) status = -1

      wcs%X%refPixels      = 0.0
      wcs%Y%refPixels      = 0.0
      wcs%X%pixelSizeWorld = 0.0
      wcs%Y%pixelSizeWorld = 0.0
      wcs%X%refWorld       = 0.0
      wcs%Y%refWorld       = 0.0
      wcs%X%ctype          = ''
      wcs%Y%ctype          = ''

  return
    end if

    wcs%X%refPixels      = real32Attribute(inArray, 'CRPIX1')
    wcs%Y%refPixels      = real32Attribute(inArray, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(inArray, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(inArray, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(inArray, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(inArray, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(inArray, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(inArray, 'CTYPE2')

    if(present(status)) status = localStatus

  end subroutine getWcsInfoArray

  !---------------------------------------------------------------------
  subroutine getWcsInfoArray_old(inArray, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT),   intent(in) :: inArray
    type(wcsDataType), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    wcs%X%refPixels      = real32Attribute(inArray, 'CRPIX1')
    wcs%Y%refPixels      = real32Attribute(inArray, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(inArray, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(inArray, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(inArray, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(inArray, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(inArray, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(inArray, 'CTYPE2')

  end subroutine getWcsInfoArray_old

  !---------------------------------------------------------------------
  subroutine getWcsInfoSet(set, wcs, status)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT),    intent(in)            :: set
    type(wcsDataType), intent(out)           :: wcs
    integer,           intent(out), optional :: status

    integer :: localStatus

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    if (.not. (hasAttribute(set, 'CRPIX1')&
         .and. hasAttribute(set, 'CRPIX2')&
         .and. hasAttribute(set, 'CDELT1')&
         .and. hasAttribute(set, 'CDELT2')&
         .and. hasAttribute(set, 'CRVAL1')&
         .and. hasAttribute(set, 'CRVAL2')&
         .and. hasAttribute(set, 'CTYPE1')&
         .and. hasAttribute(set, 'CTYPE2'))) then
      if(present(status)) status = -1

      wcs%X%refPixels      = 0.0
      wcs%Y%refPixels      = 0.0
      wcs%X%pixelSizeWorld = 0.0
      wcs%Y%pixelSizeWorld = 0.0
      wcs%X%refWorld       = 0.0
      wcs%Y%refWorld       = 0.0
      wcs%X%ctype          = ''
      wcs%Y%ctype          = ''

  return
    end if

    wcs%X%refPixels      = real32Attribute(set, 'CRPIX1')
    wcs%Y%refPixels      = real32Attribute(set, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(set, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(set, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(set, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(set, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(set, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(set, 'CTYPE2')

    if(present(status)) status = localStatus

  end subroutine getWcsInfoSet

  !---------------------------------------------------------------------
  subroutine getWcsInfo_set_old(set, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT),    intent(in)            :: set
    type(wcsDataType), intent(out)           :: wcs

    !-------------------------------------------------------------------
    ! Start:

    wcs%X%refPixels      = real32Attribute(set, 'CRPIX1')
    wcs%Y%refPixels      = real32Attribute(set, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(set, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(set, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(set, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(set, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(set, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(set, 'CTYPE2')

  end subroutine getWcsInfo_set_old

  !---------------------------------------------------------------------
  subroutine getWcsInfo_table(tab, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    type(wcsDataType), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    wcs%X%refPixels      = real32Attribute(tab, 'CRPIX1')
    wcs%Y%refPixels      = real32Attribute(tab, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(tab, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(tab, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(tab, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(tab, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(tab, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(tab, 'CTYPE2')

  end subroutine getWcsInfo_table

!  !---------------------------------------------------------------------
!  subroutine getWcsInfoSourceListSet(setName, wcs)
!  !---------------------------------------------------------------------
!!*** These kwds should surely be stored as proper column kwds?
!
!    implicit none
!
!    character(*), intent(in) :: setName
!    type(wcsDataType), intent(out) :: wcs
!
!    type(DataSetT) :: set
!
!    !-------------------------------------------------------------------
!    ! Start:
!
!    set = dataSet(setName, READ)
!
!    wcs%X%refPixels      = real32Attribute(set, 'TCRPIX1')
!    wcs%Y%refPixels      = real32Attribute(set, 'TCRPIX2')
!    wcs%X%pixelSizeWorld = real32Attribute(set, 'TCDELT1')
!    wcs%Y%pixelSizeWorld = real32Attribute(set, 'TCDELT2')
!    wcs%X%refWorld       = real32Attribute(set, 'TCRVAL1')
!    wcs%Y%refWorld       = real32Attribute(set, 'TCRVAL2')
!    wcs%X%ctype          = stringAttribute(set, 'TCTYPE1')
!    wcs%Y%ctype          = stringAttribute(set, 'TCTYPE2')
!
!    call release(set)
!
!  end subroutine getWcsInfoSourceListSet

  !---------------------------------------------------------------------
  subroutine easifyWcsType_scalar(ctypeX, ctypeY, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctypeX, ctypeY
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctypeX) .eq. 'DETX' .and. trim(ctypeY) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctypeX).eq.'RA---TAN' .and. trim(ctypeY) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsType_scalar

  !---------------------------------------------------------------------
  subroutine easifyWcsType_vector(ctype, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctype(2)
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctype(1)) .eq. 'DETX' .and. trim(ctype(2)) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctype(1)).eq.'RA---TAN' .and. trim(ctype(2)) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsType_vector

  !---------------------------------------------------------------------
  subroutine pixelsToTanScalar(xPixels, yPixels, wcs, tanX, tanY)
  !---------------------------------------------------------------------

    implicit none

    real(single),      intent(in)  :: xPixels, yPixels
    type(wcsDataType), intent(in)  :: wcs
    real(double),      intent(out) :: tanX, tanY

    !-------------------------------------------------------------------
    ! Start

    tanX = (xPixels - wcs%x%refPixels) * abs(wcs%x%pixelSizeWorld) * PiOn180
    tanY = (yPixels - wcs%y%refPixels) * abs(wcs%y%pixelSizeWorld) * PiOn180

  end subroutine pixelsToTanScalar

  !---------------------------------------------------------------------
  subroutine pixelsToTanVector(xPixels, yPixels, wcs, tanX, tanY)
  !---------------------------------------------------------------------

    implicit none

    real(single),      intent(in)  :: xPixels(:), yPixels(size(xPixels))
    type(wcsDataType), intent(in)  :: wcs
    real(double),      intent(out) :: tanX(size(xPixels)), tanY(size(xPixels))

    !-------------------------------------------------------------------
    ! Start

    tanX = (xPixels - wcs%x%refPixels) * abs(wcs%x%pixelSizeWorld) * PiOn180
    tanY = (yPixels - wcs%y%refPixels) * abs(wcs%y%pixelSizeWorld) * PiOn180

  end subroutine pixelsToTanVector

  !---------------------------------------------------------------------
  subroutine tanToPixelsScalar(wcs, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------

    implicit none

    real(double),      intent(in)  :: tanX, tanY
    type(wcsDataType), intent(in)  :: wcs
    real(single),      intent(out) :: xPixels, yPixels

    !-------------------------------------------------------------------
    ! Start

    xPixels = tanX / abs(wcs%x%pixelSizeWorld) / PiOn180 + wcs%x%refPixels 
    yPixels = tanY / abs(wcs%y%pixelSizeWorld) / PiOn180 + wcs%y%refPixels 

  end subroutine tanToPixelsScalar

  !---------------------------------------------------------------------
  subroutine tanToPixelsVector(wcs, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------

    implicit none

    real(double),      intent(in)  :: tanX(:), tanY(size(tanX))
    type(wcsDataType), intent(in)  :: wcs
    real(single),      intent(out) :: xPixels(size(tanX)), yPixels(size(tanX))

    !-------------------------------------------------------------------
    ! Start

    xPixels = tanX / abs(wcs%x%pixelSizeWorld) / PiOn180 + wcs%x%refPixels 
    yPixels = tanY / abs(wcs%y%pixelSizeWorld) / PiOn180 + wcs%y%refPixels 

  end subroutine tanToPixelsVector

  !---------------------------------------------------------------------
  function tanToPixelsAxisScalar(wcsAxis, tan) result(pixels)
  !---------------------------------------------------------------------

    implicit none

    real(double),      intent(in)  :: tan
    type(wcsAxisType), intent(in)  :: wcsAxis
    real(single)                   :: pixels

    !-------------------------------------------------------------------
    ! Start

    pixels = tan / abs(wcsAxis%pixelSizeWorld) / PiOn180 + wcsAxis%refPixels 

  end function tanToPixelsAxisScalar

  !---------------------------------------------------------------------
  function tanToPixelsAxisVector(wcsAxis, tan) result(pixels)
  !---------------------------------------------------------------------

    implicit none

    real(double),      intent(in)  :: tan(:)
    type(wcsAxisType), intent(in)  :: wcsAxis
    real(single)                   :: pixels(size(tan))

    !-------------------------------------------------------------------
    ! Start

    pixels = tan / abs(wcsAxis%pixelSizeWorld) / PiOn180 + wcsAxis%refPixels 

  end function tanToPixelsAxisVector

  !---------------------------------------------------------------------
  function tanToPixelsAxisArray(wcsAxis, tan) result(pixels)
  !---------------------------------------------------------------------

    implicit none

    real(double),      intent(in)  :: tan(:,:)
    type(wcsAxisType), intent(in)  :: wcsAxis
    real(single)                   :: pixels(size(tan,1),size(tan,2))

    !-------------------------------------------------------------------
    ! Start

    pixels = tan / abs(wcsAxis%pixelSizeWorld) / PiOn180 + wcsAxis%refPixels 

  end function tanToPixelsAxisArray

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module wcs_aux
