! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: bkgcountsmap_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Makes an image of the background contained in an nxn sliding box.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module bkgCountsMap_mod

  use dal
  use param  
  use errorHandling

  use dal_aux
  use save_image

  implicit none 

  private 

  ! Command-line parameters:
  character(255) ::&
    expMapName,&
    nonVigMapName,&
    bkgRateMapName,&
    bkgCountsMapName,&
    tempImageName
  integer(int16) ::&
    boxHalfSize
  real(single) ::&
    fractionOfBkgVignetted

  ! Other globals:

  public bkgCountsMap

contains 

  !---------------------------------------------------------------------
  subroutine bkgCountsMap
  !---------------------------------------------------------------------
  ! This sums a vignetted and a non-vignetted exposure map in the ratios appropriate to the background conditions and convolves the result with an nxn box. The background rate map is then multiplied by the convolved exposure map, to give a map of the background counts to be expected in an nxn sliding box.
  !
  ! Note that the input background rate map values should be in counts/s/pixel, and should extend beyond the edges of the non-zero sections of the exposure map by at least n-1 in all directions.

    implicit none

    integer(int16) :: xSize, ySize
    real(single), pointer :: expMap(:,:), nonVigMap(:,:), bkgRateMap(:,:)
    real(single), allocatable :: bkgCountsImage(:,:)

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    call readArrayData(expMapName, expMap)
    xSize = size(expMap, 1)
    ySize = size(expMap, 2)
    call readArrayData(nonVigMapName, nonVigMap)
    call readArrayData(bkgRateMapName, bkgRateMap)
    if ( xSize .ne. size(bkgRateMap, 1)&
    .or. ySize .ne. size(bkgRateMap, 2)&
    .or. xSize .ne. size(nonVigMap,  1)&
    .or. ySize .ne. size(nonVigMap,  2))&
call error('','')

    expMap = nonVigMap * (1.0 - fractionOfBkgVignetted)&
      + expMap * fractionOfBkgVignetted

    expMap = directBoxConvolve(expMap, boxHalfSize, boxHalfSize)
    allocate(bkgCountsImage(xSize, ySize))
    bkgCountsImage = bkgRateMap * expMap

    call saveImage(bkgCountsImage, expMapName, bkgCountsMapName, tempImageName)

    deallocate(expMap, bkgRateMap)

  end subroutine bkgCountsMap

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    boxHalfSize = intParameter('boxhalfsize')
    fractionOfBkgVignetted = realParameter('fracvignetted')
    expMapName       = stringParameter('expimageset')
    nonVigMapName    = stringParameter('nonvigimageset')
    bkgRateMapName   = stringParameter('bkgimageset')
    bkgCountsMapName = stringParameter('bkgcountsimageset')
    tempImageName    = stringParameter('tempset')

  end subroutine readParameters

  !---------------------------------------------------------------------
  function directBoxConvolve(inImage, boxHalfXSize, boxHalfYSize)&
  result(outImage)
  !---------------------------------------------------------------------

    implicit none

    real(single),   intent(in) :: inImage(:,:)
    integer(int16), intent(in) :: boxHalfXSize, boxHalfYSize
    real(single)               :: outImage(size(inImage,1), size(inImage,2))

    integer(int16) :: xSize, ySize, xi, yi, xiStart, xiFinis, yiStart, yiFinis

    !-------------------------------------------------------------------
    ! Start

    xSize = size(inImage, 1)
    ySize = size(inImage, 2)
    outImage = 0.0 ! default

    do xi = 1, xSize
      xiStart = max(1_int16, xi - boxHalfXSize)
      xiFinis = min(xSize,   xi + boxHalfXSize)
      do yi = 1, ySize
        yiStart = max(1_int16, yi - boxHalfYSize)
        yiFinis = min(ySize,   yi + boxHalfYSize)

        outImage(xi, yi) = sum(inImage(xiStart:xiFinis,yiStart:yiFinis))
      end do
    end do

  end function directBoxConvolve

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module bkgCountsMap_mod
