! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: sensitivity_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Creates a map of sensitivity in counts s^-1, for a given exposure and energy band.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module sensitivity_mod

  use dal
  use param  
  use errorHandling

  use dal_aux
  use array_utils
  use math_utils
  use save_image
  use detection_stats
  use ssc_misc
use dump

  implicit none 

  private 

  ! Command-line parameters:
  real(single), allocatable ::&
    srcBandRateFraction(:)
  real(single) ::&
    likelihoodCutoff
  character(255), allocatable ::&
    bkgCountsMapName(:),&
    psfBoxFracMapName(:),&
    exposureMapName(:),&
    sensitivityMapName(:)
!  character(255) ::&
!    tempImageName

  ! Other globals:
!  integer(int16) ::&
!    numBands
  integer(int16) ::&
    xiStart,&
    xiFinis,&
    xSize,&
    yiStart,&
    yiFinis,&
    ySize
!  real(single), allocatable :: xValuesForIncomGamma(:), aMultForIncomGamma(:)
!  real(single) :: targetLikelihood

  public sensitivity

contains 

  !---------------------------------------------------------------------
  subroutine sensitivity
  !---------------------------------------------------------------------
  ! The way this works is as follows.
  !
  ! The sensitivity of source-detection is set by the likelihood cutoff,
  ! ie the likelihood value below which putative sources are discarded.
  ! There is a fixed relationship between this cutoff, the background
  ! level and the minimum source counts; the exact nature of this relation
  ! depends on the statistics associated with the image and the detection
  ! scheme.
  !
  ! In the XMM case sources are detected by eboxdetect (in map mode),
  ! whereas emldetect (more or less) just performs an accurate
  ! characterisation of sources already detected. Hence the eboxdetect
  ! likelihood cutoff and sliding-box statistics are relevant to a
  ! determination of the minimum detectable source counts.
  !
  ! A complicating factor is that detection can be carried out in parallel
  ! over several energy bands. In this case the likelihood cutoff is
  ! applied not to each band individually but to the summed likelihoods
  ! from all the detection bands. Sensitivity in any given band will in
  ! this case depend on the spectrum of the source. Thus the present task
  ! allows the user to input the source 'spectrum', actually a list of
  ! relative count rates in each of the submitted bands.
  !
  ! The first step in the algorithm is to convert this vector of
  ! relative count rates into a stack of images giving relative counts
  ! expected in an nxn box of image pixels. There is 1 image per
  ! band. These are created by calculating what fraction of the PSF falls
  ! within a centred nxn box at each pixel.
  !
  ! For each part of the image (or for a subset of samples), an iterative
  ! procedure is then followed to determine the strength of source over all
  ! bands that would result in a net likelihood exactly equal to the cutoff
  ! value, for the given background counts. This source strength is
  ! translated back into total-PSF values then converted to a count rate
  ! by dividing by the exposure map.
  !---------------------------------------------------------------------

    implicit none

    integer(int16) :: numBands, bandNum, xi, yi
    real(single), pointer ::&
      bkgBoxCounts(:,:,:), srcBoxExposure(:,:,:),&
      srcBoxSpectrum(:,:,:), detectableSrcBoxCts(:,:,:)
!, sensitivityMap(:,:,:)
    real(single), allocatable ::&
sensitivityMap(:,:),&
!      bkgCountMaps(:,:,:), srcCountMaps(:,:,:),&
!      detectableSrcCounts(:), 
detectableSrcBoxUncert(:), tempDetectableSrcBoxCts(:)
    integer :: foreignStatus, numNonZeroPixels, pixelNum
    real(single) :: preconditioner
!real :: q
!character(20) :: dumpFileName

    !-------------------------------------------------------------------
    ! Start

!*** input _flux_ ratios, then convert to _rate_ ratios??

    call readParameters

    numBands = size(srcBandRateFraction)

    call checkAndGetImageSize ! Checks that all images are the same
    ! dimensions, and sets globals xiStart, xiFinis, xSize, yiStart,
    ! yiFinis & ySize.

!*** should also check for box dimension keywords and complain if any two are not the same.

    allocate(bkgBoxCounts(numBands, xiStart:xiFinis, yiStart:yiFinis))
    allocate(srcBoxExposure(numBands, xiStart:xiFinis, yiStart:yiFinis))
!    allocate(sensitivityMap(numBands, xiStart:xiFinis, yiStart:yiFinis))
allocate(sensitivityMap(xSize, ySize))

    srcBoxSpectrum      => srcBoxExposure
    detectableSrcBoxCts => bkgBoxCounts
    ! This is done just to have some more informative names for the various calculated values without actually needing a new storage area for each.

!    allocate(bkgCountMaps(numBands, xiStart:xiFinis, yiStart:yiFinis))
!    allocate(srcCountMaps(numBands, xiStart:xiFinis, yiStart:yiFinis))
!    allocate(detectableCounts(numBands), detectableCountsUncert(numBands))
allocate(tempDetectableSrcBoxCts(numBands))
allocate(detectableSrcBoxUncert(numBands))

    do bandNum = 1, numBands
      call loadImage(bkgCountsMapName(bandNum), bkgBoxCounts(bandNum,:,:))
      call loadImage(psfBoxFracMapName(bandNum)&
        , srcBoxExposure(bandNum,:,:))

      srcBoxSpectrum(bandNum,:,:) = srcBoxExposure(bandNum,:,:)&
        * srcBandRateFraction(bandNum)
    end do

    numNonZeroPixels = count(srcBoxSpectrum > 0.0) / real(numBands) ! approx.
    pixelNum = 0
    preconditioner = 1.0
    do xi = xiStart, xiFinis
      do yi = yiStart, yiFinis
        if (any(srcBoxSpectrum(:, xi, yi) <= 0.0&
        .or.    bkgBoxCounts(    :, xi, yi) <= 0.0))&
      cycle

        pixelNum = pixelNum + 1
        call printEvery(pixelNum, numNonZeroPixels, fraction=0.1)

        call minDetPoissonCounts(bkgBoxCounts(:, xi, yi)&
          , likelihoodCutoff, tempDetectableSrcBoxCts&
          , detectableSrcBoxUncert&
          , srcBoxSpectrum(:, xi, yi) * preconditioner, foreignStatus)
        ! in ssclib/detection_stats.

        preconditioner = tempDetectableSrcBoxCts(1) / srcBoxSpectrum(1, xi, yi)
        detectableSrcBoxCts(:, xi, yi) = tempDetectableSrcBoxCts

        if (foreignStatus < 0) then
print *, 'foreignStatus', foreignStatus
call error('badMinDetPoissonCountsStatus',"The subroutine &
&minDetPoissonCounts returned with a bad status.")
          call sensCalcErrorHandler(foreignStatus&
            , detectableSrcBoxCts(:, xi, yi))
        else if (foreignStatus > 0) then
          call warning('nonZeroStatusSensCalc','')
        end if
      end do
    end do

!    sensitivityMap = 0.0
    do bandNum = 1, numBands
      call loadImage(psfBoxFracMapName(bandNum)&
        , srcBoxExposure(bandNum,:,:))

sensitivityMap = 0.0
      where(srcBoxExposure(bandNum,:,:) > 0.0)&
!        sensitivityMap(bandNum,:,:) = detectableSrcBoxCts(bandNum,:,:)&
sensitivityMap(xiStart:xiFinis, yiStart:yiFinis)&
  = detectableSrcBoxCts(bandNum,:,:)&
          / srcBoxExposure(bandNum,:,:) ! sens map gives the minimum
        ! detectable rate in each band.

!      call saveImage(sensitivityMap(bandNum,:,:), exposureMapName(bandNum)&
call saveImage(sensitivityMap, exposureMapName(bandNum)&
        , sensitivityMapName(bandNum), 'tempImage.ds') ! in ssclib/save_image.
    end do

    deallocate(bkgBoxCounts, srcBoxExposure, sensitivityMap)

  end subroutine sensitivity

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer(int16) :: numBands, bandNum

    !-------------------------------------------------------------------
    ! Start

    numBands = parameterCount('weights')
    if (numBands <= 0)&
call error('tooFewWeights',"You must supply 1 or more --weights.")

    if (numBands .ne. parameterCount('boxbkgmapsets'))&
call error('wrongNumBoxBackgroundSets',"You must supply the same number &
&(and order) of elements in the list --boxbkgmapsets as in the list &
&--weights.")

    if (numBands .ne. parameterCount('psfexpmapsets'))&
call error('wrongNumBoxExposureSets',"You must supply the same number &
&(and order) of elements in the list --psfexpmapsets as in the list &
&--weights.")

    if (numBands .ne. parameterCount('expmapsets'))&
call error('wrongNumExpImageSets',"You must supply the same number &
&(and order) of elements in the list --expmapsets as in the list &
&--weights.")

    if (numBands .ne. parameterCount('sensmapsets'))&
call error('wrongNumSensMapSets',"You must supply the same number &
&(and order) of elements in the list --sensmapsets as in the list &
&--weights.")

    allocate(bkgCountsMapName(numBands), psfBoxFracMapName(numBands))
    allocate(exposureMapName(numBands), sensitivityMapName(numBands))
    allocate(srcBandRateFraction(numBands))

    do bandNum = 1, numBands
      srcBandRateFraction(bandNum) = realParameter('weights'&
        , bandNum-1)
      bkgCountsMapName(bandNum)    = stringParameter('boxbkgmapsets'&
        , bandNum-1)
      psfBoxFracMapName(bandNum)   = stringParameter('psfexpmapsets'&
        , bandNum-1)
      exposureMapName(bandNum)     = stringParameter('expmapsets'&
        , bandNum-1)
      sensitivityMapName(bandNum)  = stringParameter('sensmapsets'&
        , bandNum-1)
    end do

    likelihoodCutoff = realParameter('likemin')
!    tempImageName = stringParameter('tempset')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine checkAndGetImageSize
  !---------------------------------------------------------------------
  ! (i) Check that all the images have the same dimensions; (ii) get the size of the exposure map minus its blank borders.


!**** can we use ssclib/wcs_aux:checkImageEquivalence??


    implicit none

    integer(int16) :: numBands, bandNum
    real(single), pointer :: exposureMap(:,:)

    !-------------------------------------------------------------------
    ! Start

    numBands = size(srcBandRateFraction)

    if (imageSizesUnequal(bkgCountsMapName(1), psfBoxFracMapName(1)))&
call error('imageDimensionsUnequal',"The dimensions of the images in &
&--boxbkgmapsets and --psfexpmapsets must be the same.")

    if (imageSizesUnequal(bkgCountsMapName(1), exposureMapName(1)))&
call error('imageDimensionsUnequal',"The dimensions of the images in &
&--boxbkgmapsets and --expmapsets must be the same.")

    do bandNum = 2, numBands
      if (imageSizesUnequal(bkgCountsMapName(1), bkgCountsMapName(bandNum)))&
call error('imageDimensionsUnequal',"The images in the list --boxbkgmapsets &
&must all have the same size.")

      if (imageSizesUnequal(bkgCountsMapName(1), psfBoxFracMapName(bandNum)))&
call error('imageDimensionsUnequal',"The images in the list --psfexpmapsets &
&must all have the same size.")

      if (imageSizesUnequal(bkgCountsMapName(1), exposureMapName(bandNum)))&
call error('imageDimensionsUnequal',"The images in the list --expmapsets &
&must all have the same size.")
    end do

    call readArrayData(exposureMapName(1), exposureMap) ! in fits_utils
    xSize = size(exposureMap, 1)
    ySize = size(exposureMap, 2)
!    call maskArrayLimits(logical(exposureMap > 0.0, bool)&
    call maskArrayLimits((exposureMap > 0.0)&
      , xiStart, xiFinis, yiStart, yiFinis) ! in array_utils
    deallocate(exposureMap)

  end subroutine checkAndGetImageSize

  !---------------------------------------------------------------------
  logical function imageSizesUnequal(imageNameA, imageNameB)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageNameA, imageNameB

    type(DataSetT) :: setA, setB
    integer, pointer :: dimsA(:), dimsB(:)

    !-------------------------------------------------------------------
    ! Start

    imageSizesUnequal = .true. ! default

    setA = dataSet(imageNameA, READ)
    dimsA => dimensions(array(setA, 0))
    setB = dataSet(imageNameB, READ)
    dimsB => dimensions(array(setB, 0))

    if (size(dimsA) .ne. 2 .or. size(dimsB) .ne. 2) then
      call release(setA)
      call release(setB)
  return
    end if
    if (dimsA(0) .ne. dimsB(0) .or. dimsA(1) .ne. dimsB(1)) then
      call release(setA)
      call release(setB)
  return
    end if

    call release(setA)
    call release(setB)

    imageSizesUnequal = .false.

  end function imageSizesUnequal

  !---------------------------------------------------------------------
  subroutine loadImage(imageSetName, image)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), intent(out) :: image(:,:)

    real(single), pointer :: arrayPointer(:,:)

    !-------------------------------------------------------------------
    ! Start

    call readArrayData(imageSetName, arrayPointer)
    ! in fits_utils

!    if  (size(arrayPointer, 1) .ne. size(image, 1)&
!    .or. size(arrayPointer, 2) .ne. size(image, 2))&
!call error('bug0','') ! because this should have already been checked.

    image = arrayPointer(xiStart:xiFinis, yiStart:yiFinis)
    deallocate(arrayPointer)

  end subroutine loadImage

  !---------------------------------------------------------------------
  subroutine sensCalcErrorHandler(status, values)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: status
    real(single), intent(out) :: values(:)

    !-------------------------------------------------------------------
    ! Start

    select case(status)
    case(-1)
call error('bug0',"Please contact the code developer!")
    case(-2)
call error('bug1',"Please contact the code developer!")
    case(-3)
call error('bug2',"Please contact the code developer!")
    case(-4)
      call warning('negativeBackground',"The subroutine minDetPoissonCounts &
        &found negative values in some of the --boxbkgmapsets images.")
      values = 0.0
    case(-5)
      call warning('negativeSrcCounts','')
      values = 0.0
    case default
print *, 'status=', status
call error('badErrorStatus',"Please contact the code developer!")
    end select

  end subroutine sensCalcErrorHandler

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module sensitivity_mod
